<?php

declare(strict_types=1);

namespace Drupal\c2pa_sign_aws_kms\EventSubscriber;

use Drupal\c2pa_sign\Event\CertificateValidateEvent;
use Drupal\c2pa_sign\Event\CreateToolEvent;
use Drupal\c2pa_sign\Event\PreSignEvent;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Extension\ModuleExtensionList;
use Drupal\Core\Utility\Error;
use phpseclib3\Crypt\PublicKeyLoader;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpKernel\Event\RequestEvent;
use Symfony\Component\HttpKernel\Event\ResponseEvent;
use Symfony\Component\HttpKernel\KernelEvents;

/**
 * @todo Add description for this subscriber.
 */
final class C2paSignAwsKmsSubscriber implements EventSubscriberInterface {

  private \Drupal\Core\Config\ImmutableConfig $config;

  /**
   * @var \Drupal\Core\Extension\ModuleExtensionList
   */
  private ModuleExtensionList $moduleExtensionlist;

  private \Drupal\Core\Config\ImmutableConfig $c2paSignConfig;

  /**
   * Constructs a C2paSignAwsKmsSubscriber object.
   */
  public function __construct(
    private readonly ConfigFactoryInterface $configFactory,
    ModuleExtensionList $moduleExtensionList,
  ) {
    $this->config = $this->configFactory->get('c2pa_sign_aws_kms.settings');
    $this->c2paSignConfig = $this->configFactory->get('c2pa_sign.settings');
    $this->moduleExtensionlist = $moduleExtensionList;

    $true = TRUE;
  }

  /**
   * Check for certificates that weren't validated because of a missing or empty
   * key file and validate them if we can find a KMS key to match.
   *
   * @param \Drupal\c2pa_sign\Event\CertificateValidateEvent $validateEvent
   *
   * @return void
   */
  public function onCertificateValidateEvent(CertificateValidateEvent $validateEvent) {
    // check to make sure the only reason this cert was rejected is because the
    // private key file was missing
    $certFile = $validateEvent->getCertificate();
    if (!empty($certFile->invalid_reasons)) {
      // reasons that are possible to overcome
      $surmountableIssues = [
        'MISSING_KEY_FILE',
        'KEY_FILE_EMPTY',
      ];
      foreach ($certFile->invalid_reasons AS $reason) {
        if (array_search($reason, $surmountableIssues) === FALSE) {
          // there is an insurmountable reason
          return;
        }
      }
    }
    // we know this certificate might have only failed because the key file was
    // missing or empty
    try {
      $key_options = c2pa_sign_aws_kms_key_options(TRUE);
    }
    catch (Exception $e) {
      $logger = \Drupal::logger('c2pa_sign_aws_kms');
      Error::logException($logger, $e);
    }
    if (empty($key_options)) {
      return;
    }
    $kmsClient = c2pa_sign_aws_kms_create_client();
    $certPublicKey = PublicKeyLoader::load($certFile->decodedCert['tbsCertificate']['subjectPublicKeyInfo']['subjectPublicKey']);
    $fingerprintCert = $certPublicKey->getFingerprint();
    if (!empty($key_options)) {
      // there are KMS options returned, check them all against this cert to see if they match
      foreach ($key_options AS $keyId => $key) {
        // get the corresponding public key
        $kmsPublickey = $kmsClient->getPublicKey(['KeyId' => $keyId]);
        $kmsPublicKeyBits = $kmsPublickey->get('PublicKey');
        // and load it into a PHPSecLibs PublicKey object
        $kmsPublicKeyLoaded = PublicKeyLoader::load($kmsPublicKeyBits);

        // get the fingerprint of the KMS key to compare against the
        // certificate's fingerprint
        $fingerprintKmsKey = $kmsPublicKeyLoaded->getFingerprint();
        if ($fingerprintKmsKey = $fingerprintCert) {
          // the public keys match, this is a valid KMS key for the given cert
          // add the KMS Key ID or Alias to the cert object so ti can be easily
          // referenced
          $certFile->kmsId = $keyId;
          // this cert file is now valid
          $validateEvent->isValid(TRUE);
          // there are no longer any invalid reasons
          $certFile->invalid_reasons = [];
          // change the key_uri to reference KMS
          $certFile->key_uri = 'AWS KMS(' . $keyId . ')';
          $certFile->kmsKey = $key;
          // since we have matched this key with the cert we can skip trying to
          // match others
          return;
        }
      }
    }
  }

  public function onPreSignEvent(PreSignEvent $event) {
    // check to see if a KMS usage is configured
    if (!($this->config->get('aws_kms')['use_aws_kms'] ?? FALSE)) {
      return;
    }
    $settings = $this->config->get('aws_kms');
    $cert_file_directory = $this->c2paSignConfig->get('certificate_file_directory');
    $certs = c2pa_sign_get_certs($cert_file_directory);

    // check to make sure we have an appropriate key in KMS and it is selected
    $cert = NULL;
    if (!empty($certs['certificates'])) {
      foreach ($certs['certificates'] AS $thisCert) {
        // check to get the cert that is configured
        if (isset($thisCert->kmsId) && $thisCert->kmsId == $settings['key'] ) {
          $cert = $thisCert;
        }
      }
    }
    if (empty($cert)) {
      // there was no valid certificate with a matcing KMS key, do nothing
      return;
    }

    // configure the signer to use the remote signer
    $signer = $event->getSigner();
    $modulePath = DRUPAL_ROOT . '/' .  $this->moduleExtensionlist->getPath('c2pa_sign_aws_kms');
    // set the remote signer to the shell script which will execute a drush command
    $signer->setRemoteSignerPath($modulePath . '/c2pa_sign_aws_kms.sh');
    $signer->setCertDetails($cert);
    $signer->setCertFilePath($cert->uri);
    $signer->setKeyFilePath('');
    // everything should now be configured for the Signer to use the remote
    // signer executable (shell script)
  }

  public function onCreateToolEvent(CreateToolEvent $event) {
    $tool = $event->getTool();
    $extra_vars = $this->config->get('aws_kms')['environment_variables'] ?? '';
    $env = [];
    if (!empty($extra_vars)) {
      foreach(explode(PHP_EOL, $extra_vars) AS $var) {
        [$k,$v] = explode('=', trim($var));
        // remove any quotes from the value string
        $v = trim($v, '"\'');
        $env[$k] = $v;
      }
    }
    if (!empty($env)) {
      $tool->setEnvironment($env);
    }

  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents(): array {
    return [
      CertificateValidateEvent::EVENT_NAME => ['onCertificateValidateEvent'],
      PreSignEvent::EVENT_NAME => ['onPreSignEvent'],
      CreateToolEvent::EVENT_NAME => ['onCreateToolEvent'],
    ];
  }

}
